@@ -25,15 +25,16 @@ from absl import logging
 from collections import defaultdict as ddict
 import dataclasses
 from lsmmdma.initializers import initialize
-from lsmmdma.metrics import SupervisedEvaluation
+from lsmmdma.metrics import Evaluation
 import lsmmdma.mmdma_functions as mmdma_fn
+from math import ceil
 import numpy as np
 from tenacity import retry, stop_after_attempt
 import torch
 import torch.nn as nn
 from torch.nn.parameter import Parameter
 from torch.utils import tensorboard
-from typing import Tuple, Any, Dict, DefaultDict, List
+from typing import Tuple, Any, Dict, DefaultDict, List, Union
 
 
 @dataclasses.dataclass(unsafe_hash=True)
@@ -53,6 +54,9 @@ class ModelGetterConfig:
   pca: bool = False
   n_seed: int = 1
   init: str = 'uniform'
+  use_unbiased_mmd: bool = True
+  window_size: int = -1
+  threshold: float = 1e-3
 
 
 class Model(nn.Module):
@@ -176,7 +180,8 @@ class Model(nn.Module):
     """
     loss, loss_components = loss_mmdma(
         [self.weights_first_view, self.weights_second_view],
-        first_view, second_view, self.cfg_model, self.keops, self.device)
+        first_view, second_view, self.n_sample1, self.n_sample2,
+        self.cfg_model, self.keops, self.device)
     return loss, loss_components
 
 
@@ -184,6 +189,8 @@ def loss_mmdma(
     params: torch.Tensor,
     first_view: torch.Tensor,
     second_view: torch.Tensor,
+    n_first_view: int,
+    n_second_view: int,
     cfg_model: ModelGetterConfig,
     keops: bool,
     device: torch.device
@@ -203,6 +210,8 @@ def loss_mmdma(
     second_view: torch.Tensor, second set of points. The dimension of this
       tensor is (sample x feature) if mode == 'primal' and (sample x sample)
       if mode == 'dual'.
+    n_first_view: int, number of samples in the first view.
+    n_second_view: int, number of samples in the second view.
     cfg_model: ModelGetterConfig, sets parameters for the MMDMA algorithm.
     keops: bool, whether to use keops when calculating MMD.
     device: torch.device, whether a cpu or gpu ('cuda') is used.
@@ -213,12 +222,15 @@ def loss_mmdma(
   embedding_fv = torch.matmul(first_view, params[0])
   embedding_sv = torch.matmul(second_view, params[1])
   mmd = mmdma_fn.squared_mmd(
-      embedding_fv, embedding_sv, cfg_model.sigmas, keops, device)
+      embedding_fv, embedding_sv, cfg_model.sigmas, keops,
+      cfg_model.use_unbiased_mmd, device)
   if cfg_model.mode == 'primal':
     penalty_fv = mmdma_fn.pen_primal(params[0], device)
     penalty_sv = mmdma_fn.pen_primal(params[1], device)
-    distortion_fv = mmdma_fn.dis_primal(first_view, params[0])
-    distortion_sv = mmdma_fn.dis_primal(second_view, params[1])
+    distortion_fv = mmdma_fn.dis_primal(
+        first_view, params[0], n_first_view)
+    distortion_sv = mmdma_fn.dis_primal(
+        second_view, params[1], n_second_view)
   elif cfg_model.mode == 'dual':
     penalty_fv = mmdma_fn.pen_dual(embedding_fv, params[0], device)
     penalty_sv = mmdma_fn.pen_dual(embedding_sv, params[1], device)
@@ -228,7 +240,8 @@ def loss_mmdma(
     raise ValueError('cfg_model.mode must be "dual" or "primal."')
   return (mmd + cfg_model.lambda1 * (penalty_fv + penalty_sv)
           + cfg_model.lambda2 * (distortion_fv + distortion_sv),
-          (mmd, penalty_fv, penalty_sv, distortion_fv, distortion_sv))
+          (mmd.item(), penalty_fv.item(), penalty_sv.item(),
+           distortion_fv.item(), distortion_sv.item()))
 
 
 def get_kernel(data: torch.Tensor, kernel_type='linear') -> torch.Tensor:
@@ -251,23 +264,22 @@ def get_kernel(data: torch.Tensor, kernel_type='linear') -> torch.Tensor:
 
 def save_loss_value(
     loss_tmp: float,
-    loss_components: Tuple[float, ...],
-    evaluation: Dict[str, float]
+    loss_components_tmp: List[float],
+    evaluation: Dict[str, float],
     ) -> Dict[str, float]:
   """Records loss related metrics.
 
   Arguments:
     loss_tmp: float, loss value.
-    loss_components: tuple, components of the loss.
+    loss_components_tmp: tuple, components of the loss.
     evaluation: dictionary, records the output values.
 
   Returns:
     evaluation: dictionary that records the output values.
   """
-  evaluation['loss'].append(loss_tmp.data.item())
   for i, val in enumerate(
       ['mmd', 'pen_fv', 'pen_sv', 'dis_fv', 'dis_sv']):
-    evaluation[val].append(loss_components[i].data.item())
+    evaluation[val].append(loss_components_tmp[i])
   return evaluation
 
 
@@ -285,23 +297,60 @@ def save_evaluation(
     evaluation: dictionary that records the output values.
   """
   for key, val in evaluation_tmp.items():
-    evaluation[key].append(val)
+    if key == 'no':
+      val = val if isinstance(val, int) else val.tolist()
+      evaluation[key].append(val)
+    else:
+      evaluation[key].append(val)
   return evaluation
 
 
+def stopping_criterion(i: int,
+                       cfg_model: ModelGetterConfig,
+                       evaluation_loss: DefaultDict[str, List[float]]) -> bool:
+  """Computes a stopping criterion.
+
+  The algorithm stops when the relative difference between the average of the
+  loss over two consecutive windows of size ``cfg_model.window_size`` is smaller
+  than cfg_model.threshold or if a maximum number of epochs ``cfg_model.n_iter``
+  is attained.
+
+  Args:
+    i: number of epochs.
+    cfg_model: model and algorithm parameters.
+    evaluation_loss: defaultdict where values of loss components are stored.
+
+  Returns:
+    Whether to stop the algorithm.
+  """
+  ws = cfg_model.window_size
+  if i >= cfg_model.n_iter:
+    return True
+  if ws == -1:
+    return i >= cfg_model.n_iter
+  if i < ws * 2:
+    return False
+  average_t = np.mean(evaluation_loss['loss'][-2 * ws: -ws])
+  average_tplus1 = np.mean(evaluation_loss['loss'][-ws:])
+  ratio = np.absolute(
+      average_t - average_tplus1) / np.maximum(np.absolute(average_t), 1.)
+  return ratio < cfg_model.threshold
+
+
 def _evaluate(
     i: int,
-    first_view: torch.FloatTensor,
-    second_view: torch.FloatTensor,
+    first_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
+    second_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
     model: torch.nn.Module,
-    eval_fn: SupervisedEvaluation,
-    loss: float,
-    loss_components: Tuple[float],
+    eval_fn: Evaluation,
+    loss: List[float],
+    loss_components: List[List[float]],
     evaluation_loss: DefaultDict[str, List[float]],
     evaluation_matching: DefaultDict[str, List[float]],
     embeddings_results: Tuple[np.ndarray],
     pca_results: Tuple[np.ndarray],
     cfg_model: ModelGetterConfig,
+    device: torch.device,
     summary_writer: tensorboard.SummaryWriter,
     workdir: str
     ) -> Tuple[DefaultDict[str, List[float]],
@@ -324,6 +373,7 @@ def _evaluate(
     embeddings_results: records embeddings during training.
     pca_results: records results of PCA on embeddings during training.
     cfg_model: contains the parameters of the model and algorithm.
+    device: either torch.device('cuda') or 'cpu'.
     summary_writer: summary writer for tensorboard.
     workdir: directory where the files are saved.
 
@@ -351,7 +401,8 @@ def _evaluate(
 
   # Records loss.
   if cfg_model.n_record != 0 and i % cfg_model.n_record == 0:
-    evaluation_loss = save_loss_value(loss, loss_components, evaluation_loss)
+    evaluation_loss = save_loss_value(
+        loss, loss_components, evaluation_loss)
     if workdir:
       for key, val in evaluation_loss.items():
         summary_writer.add_scalar(f'train/{key:s}', val[-1], i)
@@ -365,7 +416,10 @@ def _evaluate(
     logging.info('Train evaluation: %s.', eval_out)
     if workdir:
       for key, val in eval_out_dict.items():
-        summary_writer.add_scalar(f'train/{key:s}', val, i)
+        if key != 'no':
+          summary_writer.add_scalar(f'train/{key:s}', val, i)
+        else:
+          summary_writer.add_histogram(f'train/{key:s}', val, i)
         logging.info('Flushing TensorBoard writer.')
     evaluation_matching = save_evaluation(
         eval_out_dict, evaluation_matching)
@@ -374,12 +428,14 @@ def _evaluate(
   if ((cfg_model.n_eval != 0 and i % cfg_model.n_eval == 0)
       or (cfg_model.n_record != 0 and i % cfg_model.n_record == 0)):
     embeddings_results.append(
-        [embeddings_fv.detach().cpu(), embeddings_sv.detach().cpu()])
+        [embeddings_fv.detach().cpu().numpy(),
+         embeddings_sv.detach().cpu().numpy()])
 
   # Saves 2-dimensional representation obtained with PCA.
   if cfg_model.pca != 0 and i % cfg_model.pca == 0:
     pca_rep_fv, pca_rep_sv = mmdma_fn.pca(
-        embeddings_fv.detach().cpu(), embeddings_sv.detach().cpu())
+        embeddings_fv.detach().cpu().numpy(),
+        embeddings_sv.detach().cpu().numpy())
     pca_results.append([pca_rep_fv, pca_rep_sv])
 
   if (i == cfg_model.n_iter - 1
@@ -394,9 +450,9 @@ def _evaluate(
 
 def train_and_evaluate(
     cfg_model: ModelGetterConfig,
-    first_view: torch.Tensor,
-    second_view: torch.Tensor,
-    eval_fn: SupervisedEvaluation,
+    first_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
+    second_view: Union[torch.FloatTensor, torch.utils.data.DataLoader],
+    eval_fn: Evaluation,
     workdir: str,
     device: torch.device,
     ) -> Any:
@@ -404,11 +460,11 @@ def train_and_evaluate(
 
   Arguments:
     cfg_model: ModelGetterConfig, sets parameters for the MMDMA algorithm.
-    first_view: torch.Tensor, first view. (sample1 x feature1) if primal and
+    first_view: torch.Tensor, first view (sample1 x feature1) if primal and
       (sample1 x sample1) if dual.
     second_view: torch.Tensor, second view (sample2 x feature2) if primal and
       (sample2 x sample2) if dual.
-    eval_fn: SupervisedEvaluation dataclass, applies functions for evaluation.
+    eval_fn: Evaluation dataclass, applies functions for evaluation.
     workdir: str, path to output directory.
     device: torch.device, 'cuda' or 'cpu'.
 
@@ -455,20 +511,23 @@ def train_and_evaluate(
     optimizer = torch.optim.Adam(model.parameters(), lr=cfg_model.learning_rate)
     model.train()
 
-    # TODO(lpapaxanthos): stopping criterion.
-    for i in range(cfg_model.n_iter):
+    i = 0
+    while not stopping_criterion(i, cfg_model, evaluation_loss):
       optimizer.zero_grad()
       loss, loss_components = model(first_view, second_view)
       loss.backward()
       optimizer.step()
       logging.info('Epoch: %s., train loss: %s.', i, loss)
+      evaluation_loss['loss'].append(loss.item())
 
       if evaluation:
         out = _evaluate(i, first_view, second_view, model, eval_fn, loss,
                         loss_components, evaluation_loss,
                         evaluation_matching, embeddings_results, pca_results,
-                        cfg_model, summary_writer, workdir=inner_workdir)
+                        cfg_model, device, summary_writer,
+                        workdir=inner_workdir)
         evaluation_loss, evaluation_matching, embeddings_results, pca_results, summary_writer = out
+      i += 1
 
     return (loss, optimizer, model, evaluation_loss, evaluation_matching,
             embeddings_results, pca_results)
